﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Windows.Forms;
using Microsoft.Dynamics.Framework.UI.Extensibility;
using Microsoft.Dynamics.Framework.UI.Extensibility.WinForms;

namespace Vjeko.com.Extensibility
{
    public abstract class EventAwareControlAddInBase<T> : WinFormsControlAddInBase where T : Control
    {
        private readonly List<EventManager> _eventManagers = new List<EventManager>();

        private Delegate GetOnEventDelegate(EventInfo e, EventManager manager)
        {
            Type eventArgsType = e.EventHandlerType.GetMethod("Invoke").GetParameters()[1].ParameterType;
            MethodInfo onEventTemplate = typeof(EventManager).GetMethod("OnEvent");
            MethodInfo onEvent = onEventTemplate.MakeGenericMethod(new[] { eventArgsType });
            return onEvent.CreateDelegate(e.EventHandlerType, manager);
        }

        private void AssertInnerControlAssigned()
        {
            if (InnerControl == null)
                throw new Exception("Before subscribing to events, you must instantiate a System.Windows.Forms.Control and assign it to the InnerControl property.");
        }

        protected override Control CreateControl()
        {
            InnerControl = Activator.CreateInstance<T>();
            if (ControlAddInReady != null)
                ControlAddInReady(this, null);
            return InnerControl;
        }

        [ApplicationVisible]
        public T InnerControl { get; protected set; }

        [ApplicationVisible]
        public event GenericEventHandler OnAnyEvent;

        [ApplicationVisible]
        public event EventHandler ControlAddInReady;

        [ApplicationVisible]
        public void SubscribeToEvent(string eventName)
        {
            AssertInnerControlAssigned();
            if (_eventManagers.FirstOrDefault(m => m.EventName == eventName) != null)
                return;

            EventManager manager = new EventManager { EventName = eventName };
            manager.OnAnyEvent += OnAnyEvent;
            _eventManagers.Add(manager);

            EventInfo e = typeof(T).GetEvent(eventName);
            if (e == null)
                throw new Exception("Event " + eventName + " was not found on the type " + typeof(T).FullName);

            e.AddEventHandler(InnerControl, GetOnEventDelegate(e, manager));
        }

        [ApplicationVisible]
        public void UnsubscribeFromEvent(string eventName)
        {
            AssertInnerControlAssigned();
            EventManager manager = _eventManagers.FirstOrDefault(m => m.EventName == eventName);
            if (manager != null)
            {
                EventInfo e = typeof(T).GetEvent(eventName);
                if (e == null)
                    throw new Exception("Event " + eventName + " was not found on the type " + typeof(T).FullName);

                e.RemoveEventHandler(InnerControl, GetOnEventDelegate(e, manager));

                _eventManagers.Remove(manager);
            }
        }
    }
}
